=== modified file 'bzr'
--- bzr	2007-01-16 15:56:53 +0000
+++ bzr	2007-01-24 20:40:20 +0000
@@ -73,6 +73,14 @@
 import bzrlib.lazy_regex
 bzrlib.lazy_regex.install_lazy_compile()
 
+import bzrlib.decorators
+if ('--lsprof' in sys.argv
+    or '--lsprof-file' in sys.argv
+    or '--profile' in sys.argv):
+    bzrlib.decorators.use_pretty_decorators()
+else:
+    bzrlib.decorators.use_fast_decorators()
+
 import bzrlib.commands
 import bzrlib.trace
 

=== modified file 'bzrlib/decorators.py'
--- bzrlib/decorators.py	2006-07-28 16:05:23 +0000
+++ bzrlib/decorators.py	2007-01-25 15:54:29 +0000
@@ -17,9 +17,98 @@
 
 __all__ = ['needs_read_lock',
            'needs_write_lock',
+           'use_fast_decorators',
+           'use_pretty_decorators',
            ]
 
-def needs_read_lock(unbound):
+
+def _get_parameters(func):
+    """Recreate the parameters for a function using introspection.
+
+    :return: (function_params, calling_params)
+        function_params: is a string representing the parameters of the
+            function. (such as "a, b, c=None, d=1")
+            This is used in the function declaration.
+        calling_params: is another string representing how you would call the
+            function with the correct parameters. (such as "a, b, c=c, d=d")
+            Assuming you sued function_params in the function declaration, this
+            is the parameters to put in the function call.
+
+        For example:
+
+        def wrapper(%(function_params)s):
+            return original(%(calling_params)s)
+    """
+    # "import inspect" should stay in local scope. 'inspect' takes a long time
+    # to import the first time. And since we don't always need it, don't import
+    # it globally.
+    import inspect
+    args, varargs, varkw, defaults = inspect.getargspec(func)
+    formatted = inspect.formatargspec(args, varargs=varargs,
+                                      varkw=varkw,
+                                      defaults=defaults)
+    if defaults is None:
+        args_passed = args
+    else:
+        first_default = len(args) - len(defaults)
+        args_passed = args[:first_default]
+        for arg in args[first_default:]:
+            args_passed.append("%s=%s" % (arg, arg))
+    if varargs is not None:
+        args_passed.append('*' + varargs)
+    if varkw is not None:
+        args_passed.append('**' + varkw)
+    args_passed = ', '.join(args_passed)
+
+    return formatted[1:-1], args_passed
+
+
+def _pretty_needs_read_lock(unbound):
+    """Decorate unbound to take out and release a read lock.
+
+    This decorator can be applied to methods of any class with lock_read() and
+    unlock() methods.
+    
+    Typical usage:
+        
+    class Branch(...):
+        @needs_read_lock
+        def branch_method(self, ...):
+            stuff
+    """
+    # This compiles a function with a similar name, but wrapped with
+    # lock_read/unlock calls. We use dynamic creation, because we need the
+    # internal name of the function to be modified so that --lsprof will see
+    # the correct name.
+    # TODO: jam 20070111 Modify this template so that the generated function
+    #       has the same argument signature as the original function, which
+    #       will help commands like epydoc.
+    #       This seems possible by introspecting foo.func_defaults, and
+    #       foo.func_code.co_argcount and foo.func_code.co_varnames
+    template = """\
+def %(name)s_read_locked(%(params)s):
+    self.lock_read()
+    try:
+        return unbound(%(passed_params)s)
+    finally:
+        self.unlock()
+read_locked = %(name)s_read_locked
+"""
+    params, passed_params = _get_parameters(unbound)
+    variables = {'name':unbound.__name__,
+                 'params':params,
+                 'passed_params':passed_params,
+                }
+    func_def = template % variables
+
+    exec func_def in locals()
+
+    read_locked.__doc__ = unbound.__doc__
+    read_locked.__name__ = unbound.__name__
+    return read_locked
+
+
+def _fast_needs_read_lock(unbound):
     """Decorate unbound to take out and release a read lock.
 
     This decorator can be applied to methods of any class with lock_read() and
@@ -43,7 +132,32 @@
     return read_locked
 
 
-def needs_write_lock(unbound):
+def _pretty_needs_write_lock(unbound):
+    """Decorate unbound to take out and release a write lock."""
+    template = """\
+def %(name)s_write_locked(%(params)s):
+    self.lock_write()
+    try:
+        return unbound(%(passed_params)s)
+    finally:
+        self.unlock()
+write_locked = %(name)s_write_locked
+"""
+    params, passed_params = _get_parameters(unbound)
+    variables = {'name':unbound.__name__,
+                 'params':params,
+                 'passed_params':passed_params,
+                }
+    func_def = template % variables
+
+    exec func_def in locals()
+
+    write_locked.__doc__ = unbound.__doc__
+    write_locked.__name__ = unbound.__name__
+    return write_locked
+
+
+def _fast_needs_write_lock(unbound):
     """Decorate unbound to take out and release a write lock."""
     def write_locked(self, *args, **kwargs):
         self.lock_write()
@@ -55,3 +169,26 @@
     write_locked.__name__ = unbound.__name__
     return write_locked
 
+
+# Default is more functionality, 'bzr' the commandline will request fast
+# versions.
+needs_read_lock = _pretty_needs_read_lock
+needs_write_lock = _pretty_needs_write_lock
+
+
+def use_fast_decorators():
+    """Change the default decorators to be fast loading ones.
+
+    The alternative is to have decorators that do more work to produce
+    nice-looking decorated functions, but this slows startup time.
+    """
+    global needs_read_lock, needs_write_lock
+    needs_read_lock = _fast_needs_read_lock
+    needs_write_lock = _fast_needs_write_lock
+
+
+def use_pretty_decorators():
+    """Change the default decorators to be pretty ones."""
+    global needs_read_lock, needs_write_lock
+    needs_read_lock = _pretty_needs_read_lock
+    needs_write_lock = _pretty_needs_write_lock

=== modified file 'bzrlib/tests/test_decorators.py'
--- bzrlib/tests/test_decorators.py	2006-07-28 16:05:23 +0000
+++ bzrlib/tests/test_decorators.py	2007-01-24 17:59:50 +0000
@@ -17,20 +17,40 @@
 
 """Tests for decorator functions"""
 
-from bzrlib.decorators import needs_read_lock
+import inspect
+
+from bzrlib import decorators
+from bzrlib.decorators import needs_read_lock, needs_write_lock
 from bzrlib.tests import TestCase
 
 
 class DecoratorSample(object):
     """Sample class that uses decorators.
 
-    This doesn't actually work because the class doesn't have the 
-    requiste lock_read()/unlock() methods.
+    Log when requests go through lock_read()/unlock() or lock_write()/unlock.
     """
 
+    def __init__(self):
+        self.actions = []
+
+    def lock_read(self):
+        self.actions.append('lock_read')
+
+    def lock_write(self):
+        self.actions.append('lock_write')
+
+    def unlock(self):
+        self.actions.append('unlock')
+
     @needs_read_lock
     def frob(self):
         """Frob the sample object"""
+        self.actions.append('frob')
+
+    @needs_write_lock
+    def bank(self, bar, biz=None):
+        """Bank the sample, but using bar and biz."""
+        self.actions.append(('bank', bar, biz))
 
 
 class TestDecoratorDocs(TestCase):
@@ -39,6 +59,118 @@
     def test_read_lock_passthrough(self):
         """@needs_read_lock exposes underlying name and doc."""
         sam = DecoratorSample()
-        self.assertEqual(sam.frob.__name__, 'frob')
-        self.assertEqual(sam.frob.__doc__, 
-                'Frob the sample object')
+        self.assertEqual('frob', sam.frob.__name__)
+        self.assertEqual('Frob the sample object', sam.frob.__doc__)
+
+    def test_write_lock_passthrough(self):
+        """@needs_write_lock exposes underlying name and doc."""
+        sam = DecoratorSample()
+        self.assertEqual('bank', sam.bank.__name__)
+        self.assertEqual('Bank the sample, but using bar and biz.',
+                         sam.bank.__doc__)
+
+    def test_argument_passthrough(self):
+        """Test that arguments get passed around properly."""
+        sam = DecoratorSample()
+        sam.bank('1', biz='2')
+        self.assertEqual(['lock_write',
+                          ('bank', '1', '2'),
+                          'unlock',
+                         ], sam.actions)
+
+
+class TestPrettyDecorators(TestCase):
+    """Test that pretty decorators generate nice looking wrappers."""
+
+    def get_formatted_args(self, func):
+        """Return a nicely formatted string for the arguments to a function.
+
+        This generates something like "(foo, bar=None)".
+        """
+        return inspect.formatargspec(*inspect.getargspec(func))
+
+    def test__pretty_needs_read_lock(self):
+        """Test that _pretty_needs_read_lock generates a nice wrapper."""
+
+        @decorators._pretty_needs_read_lock
+        def my_function(foo, bar, baz=None, biz=1):
+            """Just a function that supplies several arguments."""
+
+        self.assertEqual('my_function', my_function.__name__)
+        self.assertEqual('my_function_read_locked',
+                         my_function.func_code.co_name)
+        self.assertEqual('(foo, bar, baz=None, biz=1)',
+                         self.get_formatted_args(my_function))
+        self.assertEqual('Just a function that supplies several arguments.',
+                         inspect.getdoc(my_function))
+
+    def test__fast_needs_read_lock(self):
+        """Test the output of _fast_needs_read_lock."""
+
+        @decorators._fast_needs_read_lock
+        def my_function(foo, bar, baz=None, biz=1):
+            """Just a function that supplies several arguments."""
+
+        self.assertEqual('my_function', my_function.__name__)
+        self.assertEqual('read_locked', my_function.func_code.co_name)
+        self.assertEqual('(self, *args, **kwargs)',
+                         self.get_formatted_args(my_function))
+        self.assertEqual('Just a function that supplies several arguments.',
+                         inspect.getdoc(my_function))
+
+    def test__pretty_needs_write_lock(self):
+        """Test that _pretty_needs_write_lock generates a nice wrapper."""
+
+        @decorators._pretty_needs_write_lock
+        def my_function(foo, bar, baz=None, biz=1):
+            """Just a function that supplies several arguments."""
+
+        self.assertEqual('my_function', my_function.__name__)
+        self.assertEqual('my_function_write_locked',
+                         my_function.func_code.co_name)
+        self.assertEqual('(foo, bar, baz=None, biz=1)',
+                         self.get_formatted_args(my_function))
+        self.assertEqual('Just a function that supplies several arguments.',
+                         inspect.getdoc(my_function))
+
+    def test__fast_needs_write_lock(self):
+        """Test the output of _fast_needs_write_lock."""
+
+        @decorators._fast_needs_write_lock
+        def my_function(foo, bar, baz=None, biz=1):
+            """Just a function that supplies several arguments."""
+
+        self.assertEqual('my_function', my_function.__name__)
+        self.assertEqual('write_locked', my_function.func_code.co_name)
+        self.assertEqual('(self, *args, **kwargs)',
+                         self.get_formatted_args(my_function))
+        self.assertEqual('Just a function that supplies several arguments.',
+                         inspect.getdoc(my_function))
+
+    def test_use_decorators(self):
+        """Test that you can switch the type of the decorators."""
+        cur_read = decorators.needs_read_lock
+        cur_write = decorators.needs_write_lock
+        try:
+            decorators.use_fast_decorators()
+            self.assertIs(decorators._fast_needs_read_lock,
+                          decorators.needs_read_lock)
+            self.assertIs(decorators._fast_needs_write_lock,
+                          decorators.needs_write_lock)
+
+            decorators.use_pretty_decorators()
+            self.assertIs(decorators._pretty_needs_read_lock,
+                          decorators.needs_read_lock)
+            self.assertIs(decorators._pretty_needs_write_lock,
+                          decorators.needs_write_lock)
+
+            # One more switch to make sure it wasn't just good luck that the
+            # functions pointed to the correct version
+            decorators.use_fast_decorators()
+            self.assertIs(decorators._fast_needs_read_lock,
+                          decorators.needs_read_lock)
+            self.assertIs(decorators._fast_needs_write_lock,
+                          decorators.needs_write_lock)
+        finally:
+            decorators.needs_read_lock = cur_read
+            decorators.needs_write_lock = cur_write

